// BlinkGame - Hunt for the Lost Pixels
// an SDL game in honor of the BlinkenArea crew
//
// (C) 2005-06 Rene Schickbauer, Wolfgang Dautermann
//
// See LICENSE for licensing information
//


#include "globals.h"
#include "engine.h"
#include "levelhandler.h"
#include "background.h"
#include "background2.h"
#include "drawprimitives.h"
#include "showtext.h"
#include "debug.h"
#include "tiles.h"
#include "playersprite.h"
#include "colission.h"
#include "showloading.h"
#include "gameengine.h"
#include "showpicture.h"
#include "joystick.h"
#ifdef ALLOW_ACTIONCAPTURE
#include "actioncapture.h"
#endif // ALLOW_ACTIONCAPTURE
#include "sound.h"
#include "fgobjects.h"
#include "bl_lua.h"
#include "fonthandler.h"
#include "showvideo.h"
#include "fginlay.h"
#include "bmfconvert.h"
#include "errorhandler.h"
#include "triggers.h"
#include "outputfilter.h"
#include "blending.h"

#include <math.h>
#include <string.h>

Uint32 fgXoffs = 0;
Uint32 fgYoffs = 0;
double fgXoffsOld = 0;
double fgYoffsOld = 0;
double fgXoffsTmp = 0;
double fgYoffsTmp = 0;

Uint32 spriterelx;
Uint32 spriterely;
MOVE_TYPE mtype = NO_MOVE;
double spritex, spritey, spritevx, spritevy;
Uint32 mcoltype;
bool isJumping;
bool gameRunning;
double blinkdir = BLINKSPEED;
double blinkbright = 100;
Uint32 foundPixels = 0;
char errorMessage[100];
Sint32 leftright = 0;
Sint32 updown = 0;
bool leftpressed = false;
bool rightpressed = false;
bool uppressed = false;
bool downpressed = false;
bool lockedBG;
bool hasGravity;
bool freeMovement;
bool touchFGObject;
bool firstLoop;
bool canPaintSpecialTiles;
bool allowedToExit;
bool displayPlayerCoords;
bool attracktModeRunning;
bool killPlayer;

FILE* poscapfh = 0;
#ifdef ALLOW_POSITIONCAPTURE
Uint32 poscapcount;
#endif // ALLOW_POSITIONCAPTURE
bool isAttrackMode;

void initEngine()
{
	lockedBG = false;
	hasGravity = true;
	freeMovement = false;
	allowedToExit = false;
	canPaintSpecialTiles = true;
	leftpressed = false;
	rightpressed = false;
	uppressed = false;
	downpressed = false;
	
	initPlayerSprite();
	loadMonsterSprites();
	initLevel(gamedata.player->level);
	initMonsters();
	
	spritex = lhandle.respawn_x;
	spritey = lhandle.respawn_y;
	spritevx = 0;
	spritevy = 0;
	leftright = 0;
	isJumping = false;
	gameRunning = true;
	foundPixels = 0;
	firstLoop = true;
	hasEatenPixel = false;
	resetFGInlay();
	isAttrackMode = false;
	displayPlayerCoords = false;
	killPlayer = false;
	
#ifdef ALLOW_POSITIONCAPTURE
	char poscapname[100];
	sprintf(poscapname, "poscap_level_%d.bin", gamedata.player->level);
	poscapfh = fopen(configGetPath(poscapname), "wb");
	if(!poscapfh) {
		DIE(ERROR_FILE_WRITE, configGetPath(poscapname));
	}
	poscapcount = 0;
#endif // ALLOW_POSITIONCAPTURE
	
}

void deInitEngine ()
{
	deInitLevel();
	deInitPlayerSprite();
	unloadMonsterSprites();
	
	// Test FGObjects
	deInitFGObjs();
}


void renderEngine()
{
	
	errorMessage[0] = 0;
	
	// Ask SDL for the time in milliseconds
	Uint32 tick = SDL_GetTicks();
	
	// Delay a bit if we're too fast
	while (tick <= gLastTick) {
		SDL_Delay(1);
		//sprintf(errorMessage, "TOO FAST!");
		tick = SDL_GetTicks();
	}
	
	
	
	// Physics-Engine
	while (gLastTick < tick)
	{
		//fgXoffs+= leftright * 3;
		//fgYoffs+= updown * 3;
		gLastTick += 1000 / PHYSICSFPS;
		
		// Blinking of pixels
		blinkbright += blinkdir;
		if(blinkbright < 10 || blinkbright > 250) {
			blinkdir = 0 - blinkdir;
		}
		
		// Inlay message
		calcFGInlay();
		
		// Monsters
		monsterPhysics((Uint32)spritex, (Uint32)spritey);
		if(lhandle.hasScript) {
			blLuaCall("scriptPhysics", "");
		}
		handleTriggers();
		
		// Left/Right movement with smooth acceleration/braking
		
		if(leftright > 0) {
			mtype = MOVE_RIGHT;
			spritevx += SPRITE_ACCEL;
		} else if(leftright < 0) {
			mtype = MOVE_LEFT;
			spritevx -= SPRITE_ACCEL;
		} else {
			mtype = NO_MOVE;
			if(spritevx > SPRITE_ACCEL) {
				spritevx -= SPRITE_ACCEL;
			} else if(spritevx < -SPRITE_ACCEL) {
				spritevx += SPRITE_ACCEL;
			} else {
				spritevx = 0;
			}
		}
		
		if(freeMovement) {
			if(updown > 0) {
				spritevy += SPRITE_ACCEL;
			} else if(updown < 0) {
				spritevy -= SPRITE_ACCEL;
			} else {
				if(spritevy > SPRITE_ACCEL) {
					spritevy -= SPRITE_ACCEL;
				} else if(spritevy < -SPRITE_ACCEL) {
					spritevy += SPRITE_ACCEL;
				} else {
					spritevy = 0;
				}
			}
		}
		
		if(spritevx > SPRITE_MAXSPEED) {
			spritevx = SPRITE_MAXSPEED;
		} else if(spritevx < -SPRITE_MAXSPEED) {
			spritevx = -SPRITE_MAXSPEED;
		}
		spritex += spritevx;
		
		if(spritevy > SPRITE_MAXSPEED) {
			spritevy = SPRITE_MAXSPEED;
		} else if(spritevy < -SPRITE_MAXSPEED) {
			spritevy = -SPRITE_MAXSPEED;
		}
		
		spritey += spritevy;		
		
		// Limit movement to stay within game-area
		if(spritex > (lhandle.width - 32)) {
			spritevx = 0;
			spritex = lhandle.width - 32;
		} else if(spritex < 0) {
			spritevx = 0;
			spritex = 0;
		}
		
		
		if(spritey > (lhandle.height - 32)) {
			spritevy = 0;
			spritey = lhandle.height - 32;
		} else if(spritey < 0) {
			spritevy = 0;
			spritey = 0;
		}
		
		
		// Gravity
		if(hasGravity && !freeMovement) {
			spritevy += SPRITE_GRAVITY;
		}
		
		if(isAttrackMode) {
			spritex = (double)bmfReadInt(poscapfh);
			spritey = (double)bmfReadInt(poscapfh);
			mtype = (MOVE_TYPE)bmfReadInt(poscapfh);
			if(spritex == 0 && spritey == 0) {
				gamedata.player->lives = 0;
				gameRunning = false;
				return;
			}
		}
		
#ifdef ALLOW_POSITIONCAPTURE
		if(!isAttrackMode && poscapfh) {
			bmfWriteInt(poscapfh, (Uint32)spritex);
			bmfWriteInt(poscapfh, (Uint32)spritey);
			bmfWriteInt(poscapfh, (Uint32)mtype);
			poscapcount++;
			if(poscapcount == 1700) {
				bmfWriteInt(poscapfh, (Uint32)0);
				bmfWriteInt(poscapfh, (Uint32)0);
				bmfWriteInt(poscapfh, (Uint32)0);
				fclose(poscapfh);
				poscapfh = 0;
			}
		}	
#endif // ALLOW_POSITIONCAPTURE
		
		
		handlePlayerTilesColission();
		handlePlayerFGColission();

		mcoltype = getPlayerMonsterColission((Uint32)spritex, (Uint32)spritey);
		if(mcoltype & COL_UP || mcoltype & COL_LEFT || mcoltype & COL_RIGHT || getPlayerFGKillColission() || killPlayer) {
			spritevx = 0;
			spritevy = 0;
			spritex = lhandle.respawn_x;
			spritey = lhandle.respawn_y;
			gamedata.player->lives--;
			leftpressed = false;
			rightpressed = false;
			uppressed = false;
			downpressed = false;
			firstLoop = true;
			hasEatenPixel = false;
			hasGravity = true;
			freeMovement = false;
			killPlayer = false;
			if(!showVideo("livelost.bmf", false, false)) {
				soundPlayFX(FX_KILL_PLAYER);
				showPicture("livelost.jpg", 2000);
			}
#ifdef ALLOW_ACTIONCAPTURE
			stopActionCapture();
#endif // ALLOW_ACTIONCAPTURE			
			gLastTick = SDL_GetTicks();
			return;
		}
		
		touchFGObject = false;
		
		// Check if the player has enough pixels to exit
		if(foundPixels == lhandle.numPixels) {
			allowedToExit = true;
		}
		
		// Check for "colission" with goal tile
		if((Uint32)spritex >= (lhandle.goal_x - TILESIZE + 1) && (Uint32)spritex < (lhandle.goal_x + TILESIZE) &&
		   (Uint32)spritey > (lhandle.goal_y - TILESIZE + 1) && (Uint32)spritey < (lhandle.goal_y + TILESIZE)) {
			if(allowedToExit) {
				printf("LEVEL FINISHED\n");
				gameRunning = false;
			} else {
				sprintf(errorMessage, "Level not finished!");
			}
		}
		
		handlePixelCollission((Uint32)spritex, (Uint32)spritey);
		
		
		// Calculate relative positions of sprite, tiles and background
		spriterelx = SCR_WIDTH / 2;
		spriterely = SCR_HEIGHT / 2;
		
		if(spritex < spriterelx) {
			fgXoffs = 0;
			spriterelx = (Uint32)spritex;
		} else if(spritex > (lhandle.width - spriterelx)) {
			spriterelx = SCR_WIDTH - (lhandle.width - (Uint32)spritex);
			fgXoffs = lhandle.width - SCR_WIDTH;
		} else {
			fgXoffs = (Uint32)spritex - spriterelx;
		}
		
		if(spritey < spriterely) {
			fgYoffs = 0;
			spriterely = (Uint32)spritey;
		} else if(spritey > (lhandle.height - spriterely)) {
			spriterely = SCR_HEIGHT - (lhandle.height - (Uint32)spritey);
			fgYoffs = lhandle.height - SCR_HEIGHT;
		} else {
			fgYoffs = (Uint32)spritey - spriterely;
		}
		
		
#ifdef ALLOW_SMOOTHPANNING
		// --- Start of smooth panning feature
		if(firstLoop) {
			firstLoop = false;
			fgXoffsOld = (double)fgXoffs;
			fgYoffsOld = (double)fgYoffs;
		}
		
		fgXoffsTmp = fgXoffsOld - (fgXoffsOld - (double)fgXoffs) / SMOOTHPANNING_FACTOR;
		if(fgXoffsTmp < 0.0) {
			fgXoffsTmp = 0.0;
		}
		
		fgYoffsTmp = fgYoffsOld - (fgYoffsOld - (double)fgYoffs) / SMOOTHPANNING_FACTOR;
		if(fgYoffsTmp < 0.0) {
			fgYoffsTmp = 0.0;
		}
		
		//spriterelx = spriterelx + (Uint32)(fgXoffsTmp) - fgXoffs;
		//spriterely = spriterely + ((Uint32)(fgYoffsTmp) - fgYoffs);
		
		fgXoffsOld = fgXoffsTmp;
		fgYoffsOld = fgYoffsTmp;
		
		Sint32 tmpSX = (Sint32)spritex - (Sint32)fgXoffsTmp;
		if(tmpSX < 0) {
			tmpSX = 0;
		} else if(tmpSX > SCR_WIDTH - TILESIZE) {
			tmpSX = SCR_WIDTH - TILESIZE;
		}
		spriterelx = tmpSX;
		
		Sint32 tmpSY = (Sint32)spritey - (Sint32)fgYoffsTmp;
		if(tmpSY < 0) {
			tmpSY = 0;
		} else if(tmpSY > SCR_HEIGHT - TILESIZE) {
			tmpSY = SCR_HEIGHT - TILESIZE;
		}
		spriterely = tmpSY;
		
		//printf("X: %d Y: %d\n", spriterelx, spriterely);
		
		
		fgXoffs = (Uint32)fgXoffsTmp;
		fgYoffs = (Uint32)fgYoffsTmp;
		
		// --- End of smooth panning feature
#endif // ALLOW_SMOOTHPANNING
				
	}
	
	// Lock screen if needed
	if (SDL_MUSTLOCK(gScreen))
		if (SDL_LockSurface(gScreen) < 0) 
			return;
	if(!lockedBG) {
		if(!lhandle.hasBG2) {
			drawBackground(fgXoffs / 2, fgYoffs / 2);
		} else {
			drawBackground(fgXoffs / 4, fgYoffs / 4);
			drawBackground2(fgXoffs / 2, fgYoffs / 2);
		}
	} else {
		drawBackground(fgXoffs, fgYoffs);
	}
	
	paintFGObjs(fgXoffs, fgYoffs, false);
	paintLevelTiles(fgXoffs, fgYoffs);
	paintSpecialTiles(fgXoffs, fgYoffs, (allowedToExit));
	paintLevelPixels(fgXoffs, fgYoffs, (Uint32)blinkbright);
	showMonsterSprites(fgXoffs, fgYoffs);
	showSprite(spriterelx, spriterely, mtype);
	paintFGObjs(fgXoffs, fgYoffs, true);
	
	
	// HUD ("Head-up-display")
	blend_darkenRect(3, 3, 634, 24, 0x00606060);
	
	char pixelsString[100];
	sprintf(pixelsString, "Pixels: %u (%u)", foundPixels, lhandle.numPixels);
	drawstring(20, 8, pixelsString, gScreen);
	
	char livesString[100];
	sprintf(livesString, "Lives: %u", gamedata.player->lives);
	drawstring(190, 8, livesString, gScreen);
	
	//gamedata.player->level
	char levelString[100];
	sprintf(levelString, "Level: %u", gamedata.player->level);
	drawstring(310, 8, levelString, gScreen);
	
	
	char scoreString[100];
	sprintf(scoreString, "Score: %u", gamedata.player->score);
	drawstring(440, 8, scoreString, gScreen);
	
	if(enteringCheat) {
		char cheatString[100];
		sprintf(cheatString, "CHEAT: %s", cheatCode);
		drawstring(20, 450, cheatString, gScreen);
	}
	
	
	drawstring(20, 40, errorMessage, gScreen);
	if(isAttrackMode) {
		drawstring(20, 60, "DEMO MODE", gScreen);
	}
	
	
#ifdef ALLOW_POSITIONCAPTURE
	if(!isAttrackMode && poscapfh) {
		drawstring(20, 60, "Position Capture (smile, please)", gScreen);
	}
#endif // ALLOW_POSITIONCAPTURE
	
	
#ifdef DISPLAY_PLAYERCOORDS
    if(displayPlayerCoords && !enteringCheat) {
    	char coordsString[100];
    	sprintf(coordsString, "X: %u", (Uint32)spritex);
    	drawstring(20, 450, coordsString, gScreen);
    	sprintf(coordsString, "Y: %u", (Uint32)spritey);
    	drawstring(120, 450, coordsString, gScreen);
    	
    	sprintf(coordsString, "X: %u", (Uint32)((spritex/32)+1));
    	drawstring(400, 450, coordsString, gScreen);
        sprintf(coordsString, "Y: %u", (Uint32)((spritey/32)+1));
    	drawstring(500, 450, coordsString, gScreen);	
    	
    	/*
		sprintf(coordsString, "X: %u  Y: %u       X: %u Y: %u", 
			(Uint32)spritex, (Uint32)spritey,
			(Uint32)((spritex/32)+1), (Uint32)((spritey/32)+1));
		drawstring(20, 450, coordsString, gScreen);
		*/
	}
#endif // DISPLAY_PLAYERCOORDS

#ifdef SHOWFRAMERATE
	displayFPSCounter();
#endif // SHOWFRAMERATE

#ifdef SHOW_INLAY
	if(!enteringCheat && !displayPlayerCoords) {
		renderFGInlay();
	}
#endif // SHOW_INLAY

    applyOutputFilter();
	
#ifdef ALLOW_ACTIONCAPTURE
	renderActionCapture(spriterelx, spriterely);
#endif // ALLOW_ACTIONCAPTURE


	
	// Unlock screen if needed
	if (SDL_MUSTLOCK(gScreen)) 
		SDL_UnlockSurface(gScreen);
	
	// Tell SDL to update the whole screen
	SDL_Flip(gScreen);
	
}


char cheatCode[MAX_STRING_LENGTH];
bool enteringCheat = false;
void displayEngine(char* attrackModeFile)
{
	
	leftpressed = false;
	rightpressed = false;
	SDL_Event event;
	
	showLoading();
	initEngine();
	
	// Special for attrack mode
	if(attrackModeFile) {
		gLastTick = SDL_GetTicks();
		isAttrackMode = true;
		poscapfh = fopen(configGetPath(attrackModeFile), "rb");
		if(!poscapfh) {
			DIE(ERROR_FILE_READ, configGetPath(attrackModeFile));
		}
		
		// Catch all user input
		while(gameRunning) {
			// Poll for events, and handle the ones we care about.
			while (gameRunning) 
			{
				while(SDL_PollEvent(&event)) {
					switch (event.type) 
					{
						case SDL_KEYUP:
							gameRunning = false;
							attracktModeRunning = false;
							break;
						case SDL_QUIT:
							exit(0);
					}
				}
				Uint32 joymove = getJoystickMoves();
				if(joymove & JOYSTICK_JUMP || joymove & JOYSTICK_ACTION) {
					gameRunning = false;
					attracktModeRunning = false;
				}
				renderEngine();
			}
		}
		fclose(poscapfh);
		deInitEngine();
		return;
	}
	
	
	
	
	if(gamedata.player->lives > 0) {
		if(!showVideo("levelwon.bmf", false, false)) {
			soundPlayFX(FX_LEVEL_FINISHED);
		}
	}
	
	
	Uint32 cheatTick = 0;
	gLastTick = SDL_GetTicks();
	
	while (gameRunning && gamedata.player->lives > 0)
	{
		Uint32 ticks = SDL_GetTicks();
		// Render stuff
		
		// Get Joystick moves
		Uint32 joymove = getJoystickMoves();
		if(!freeMovement) {
			if((joymove & JOYSTICK_JUMP) && !isJumping && (touchFGObject || (spritevy < SPRITE_FALLSILENTTHRESHOLD))) {
				isJumping = true;
				spritevy = SPRITE_JUMPSPEED;
			}		
		}
		
		// Calculate reftright offset
		if((leftpressed && !rightpressed) || (joymove & JOYSTICK_LEFT)){
			leftright = -1;
		} else if((!leftpressed && rightpressed) || (joymove & JOYSTICK_RIGHT)){
			leftright = 1;
		} else {
			leftright = 0;
		}
		
		if(freeMovement) {
			// Calculate updown offset
			if((uppressed && !downpressed) || (joymove & JOYSTICK_UP)){
				updown = -1;
			} else if((!uppressed && downpressed) || (joymove & JOYSTICK_DOWN)){
				updown = 1;
			} else {
				updown = 0;
			}
			isJumping = 0;
		} else {
			uppressed = false;
			downpressed = false;
		}
		
		if(joymove & JOYSTICK_ACTION) {
			doFGPlayerAction((Uint32)spritex, (Uint32)spritey);
		}
		
		renderEngine();
		if(ticks > cheatTick) {
			cheatTick = SDL_GetTicks() + CHEATTIMEOUT;
			cheatCode[0] = 0;
			enteringCheat = false;
		}
		
		// Poll for events, and handle the ones we care about.
		while (SDL_PollEvent(&event) && gameRunning) 
		{
			switch (event.type) 
			{
				case SDL_KEYDOWN:
					if (event.key.keysym.sym == SDLK_LEFT) {
						//leftright -= 1;
						leftpressed = true;
					} else if (event.key.keysym.sym == SDLK_RIGHT) {
						//leftright += 1;
						rightpressed = true;
					} else if (event.key.keysym.sym == SDLK_UP) {
						if(!freeMovement) {
							if(touchFGObject || (!isJumping && spritevy < SPRITE_FALLSILENTTHRESHOLD)) {
								isJumping = true;
								spritevy = SPRITE_JUMPSPEED;
							}
						} else {
							uppressed = true;
						}
					} else if (event.key.keysym.sym == SDLK_DOWN) {
						if(freeMovement) {
							downpressed = true;
						}
					}
					
					
					break;
				case SDL_KEYUP:
					// If escape is pressed, return (and thus, quit)
					if (event.key.keysym.sym == SDLK_LEFT) {
						//leftright += 1;
						leftpressed = false;
					} else if (event.key.keysym.sym == SDLK_RIGHT) {
						rightpressed = false;
						//leftright -= 1;
					} else if (event.key.keysym.sym == SDLK_UP) {
						if(freeMovement) {
							uppressed = false;
						}
					} else if (event.key.keysym.sym == SDLK_DOWN) {
						if(freeMovement) {
							downpressed = false;
						}
#ifdef ENABLE_FULLSCREEN
					} else if ((event.key.keysym.sym == SDLK_f) && (enteringCheat == false)) {
						/* Switching between fullscreen and 
						* windowing not allowed, if we are entering a cheatcode now 
						* (otherwise a cheatcode may not contain the character 'f' */
						isFullscreen = !isFullscreen;
						if(isFullscreen) {
							gScreen = SDL_SetVideoMode(SCR_WIDTH, SCR_HEIGHT, 32, SDL_SWSURFACE | SDL_FULLSCREEN);
						} else {
							gScreen = SDL_SetVideoMode(SCR_WIDTH, SCR_HEIGHT, 32, SDL_SWSURFACE);
						}
#endif // ENABLE_FULLSCREEN
#ifdef ALLOW_ACTIONCAPTURE
					} else if (event.key.keysym.sym == SDLK_y) {
						startActionCapture();
					} else if (event.key.keysym.sym == SDLK_x) {
						stopActionCapture();
#endif // ALLOW_ACTIONCAPTURE
					} else if (!enteringCheat && event.key.keysym.sym == SDLK_SPACE) {
						doFGPlayerAction((Uint32)spritex, (Uint32)spritey);
#ifdef ALLOW_CHEATCODES
						
					} else if (event.key.keysym.sym == SDLK_TAB) {
						enteringCheat = true;
					} else if (event.key.keysym.sym == SDLK_RETURN) {
						if(strcmp(cheatCode, "nuke") == 0) {
							printf("Nuking all monsters and adding them to your score...\n");
							MONSTERS* monster = lhandle.monsters;
							while(monster) {
								if(monster->isAlive) {
									gamedata.player->score += monster->score;
									monster->isAlive = false;
									monster->isDying = true;
								}
								monster = monster->next;
							}
						} else if(strcmp(cheatCode, "skip") == 0) {
							gameRunning = false;
						} else if(strncmp(cheatCode, "goto ", 5) == 0) {
							char* instr = strchr(cheatCode, ' ');
							if(instr) {
								*instr = 0;
								instr++;
								Uint32 lnum = (Uint32)atoi(instr);
								if(lnum && lnum > 0) {
									gamedata.player->level = lnum - 1;
									gameRunning = false;
                                				}
							}
						} else if(strcmp(cheatCode, "lives") == 0) {
							gamedata.player->lives = 99;
						} else if(strcmp(cheatCode, "die") == 0) {
							gamedata.player->lives = 0;
							gameRunning = false;
						} else if(strcmp(cheatCode, "quit") == 0) {
							printf("Trying fast exit....\n");
							displaymessage(displaymessage_INFO, "Exit using cheat code", 1000);
							exit(0);
						} else if(strcmp(cheatCode, "coords") == 0) {
							displayPlayerCoords = !displayPlayerCoords;
						} else if(strcmp(cheatCode, "freemode") == 0) {
							freeMovement = !freeMovement;
						} else if(strcmp(cheatCode, "edit") == 0) {
							freeMovement = !freeMovement;
							displayPlayerCoords = !displayPlayerCoords;
							MONSTERS* monster = lhandle.monsters;
							gamedata.player->lives = 999;
							soundPlayFX(FX_KILL_PLAYER);
							showPicture("livelost.jpg", 200);
							gLastTick = SDL_GetTicks();
							while(monster) {
								if(monster->isAlive) {
									gamedata.player->score += monster->score;
									monster->isAlive = false;
									monster->isDying = true;
								}
								monster = monster->next;
							}
						}
						cheatCode[0] = 0;
						enteringCheat = false;
					} else if(enteringCheat) {
						/* Cheatcode may only contain the characters in CHEATCODECHARS */
						char tmp[2];
						sprintf(tmp, "%c", event.key.keysym.sym);
						if(strstr(CHEATCODECHARS, tmp) && strlen(cheatCode) < 30) {
							strcat(cheatCode, tmp);
						} ;
						if ((event.key.keysym.sym == SDLK_BACKSPACE)  && (strlen(cheatCode)>0)) { /* Backspace - delete last char */
							cheatCode[strlen(cheatCode)-1] = '\0' ;
						} ;
						cheatTick = SDL_GetTicks() + CHEATTIMEOUT;
#endif // ALLOW_CHEATCODES

// Handle the BOSS-Key (ESC) - fast EXIT.
#ifdef ALLOW_BOSSKEY
					} else if (event.key.keysym.sym == SDLK_ESCAPE) {
						displaymessage(displaymessage_INFO, "Boss-key (ESC) hit - fast exit", 800);
						exit(0);
#endif // ALLOW_BOSSKEY
					}
					
					
					break;
				case SDL_QUIT:
					exit(0);
			}
		}
	}

deInitEngine();
}
