// BlinkGame - Hunt for the Lost Pixels
// an SDL game in honor of the BlinkenArea crew
//
// (C) 2005-06 Rene Schickbauer, Wolfgang Dautermann
//
// See License.txt for licensing information
//


#include <stdio.h>
#include "globals.h"
#include "fgobjects.h"
#include "errorhandler.h"
#include "gameengine.h"
#include "colission.h"
#include "engine.h"
#include "bl_lua.h"
#include "sound.h"
#include "playersprite.h"
#include "convert.h"

Uint32 fgobjcnt = 0;
Uint32 fgobjgfxcnt = 0;
SDL_Surface *FGOBJ_Surface[MAX_FGOBJECTGFX];
FGOBJS fgObjs[MAX_FGOBJECTS];
FGOVERLAP fgOverlap;

void initFGObjs() {
	fgobjcnt = 0;
	fgobjgfxcnt = 0;
}

void deInitFGObjs() {
	for(Uint32 i = 0; i < fgobjgfxcnt; i++) {
		SDL_FreeSurface(FGOBJ_Surface[i]);
	}
	fgobjgfxcnt = 0;
	fgobjcnt = 0;
}

Uint32 addFGObjGFX(char *fname) {
	if(fgobjgfxcnt == MAX_FGOBJECTGFX) {
		DIE(ERROR_BOUNDARY, "MAX_FGOBJECTGFX");
	}
	
	char fullfname[MAX_FNAME_LENGTH];
	sprintf(fullfname, "%s", configGetPath(fname));
	SDL_Surface* temp = IMG_Load(fullfname);
	if(!temp) {
		DIE(ERROR_IMAGE_READ, fullfname);
	}

	FGOBJ_Surface[fgobjgfxcnt] = convertToBSSurface(temp);
	//printf("Loading FG GFX %s as %d\n", fname, fgobjgfxcnt);
	SDL_FreeSurface(temp);
	fgobjgfxcnt++;
	return(fgobjgfxcnt - 1);
}

Uint32 addFGObj(Uint32 gfxobj, Uint32 x, Uint32 y, bool isBlocking, bool isVisible, bool isOnTop, bool isKilling) {
	if(fgobjcnt == MAX_FGOBJECTS) {
		DIE(ERROR_BOUNDARY, "MAX_FGOBJECT");
	}
	if(gfxobj >= fgobjgfxcnt) {
		DIE(ERROR_INDEX_INVALID, "gfxobj");
	}
	
	fgObjs[fgobjcnt].gfxobj = gfxobj;
	fgObjs[fgobjcnt].x = x;
	fgObjs[fgobjcnt].y = y;
	fgObjs[fgobjcnt].last_y = y;
	fgObjs[fgobjcnt].isBlocking = isBlocking;
	fgObjs[fgobjcnt].isVisible = isVisible;
	fgObjs[fgobjcnt].isOnTop = isOnTop;
	fgObjs[fgobjcnt].isKilling = isKilling;
	fgObjs[fgobjcnt].isPixel = false;
	fgObjs[fgobjcnt].isElevator = false;
	
	fgobjcnt++;
	return(fgobjcnt - 1);
}

void setFGObjSetVisible(Uint32 obj, bool visible) {
	if(obj > fgobjcnt) {
		DIE(ERROR_INDEX_INVALID, "fgobj");
	}
	fgObjs[obj].isVisible = visible;
}

void setFGObjSetBlocking(Uint32 obj, bool blocking) {
	if(obj > fgobjcnt) {
		DIE(ERROR_INDEX_INVALID, "fgobj");
	}
	fgObjs[obj].isBlocking = blocking;
}

void setFGObjSetKilling(Uint32 obj, bool killing) {
	if(obj > fgobjcnt) {
		DIE(ERROR_INDEX_INVALID, "fgobj");
	}
	fgObjs[obj].isKilling = killing;
}

void setFGObjSetPixel(Uint32 obj, bool pixel) {
	if(obj > fgobjcnt) {
		DIE(ERROR_INDEX_INVALID, "fgobj");
	}
	fgObjs[obj].isPixel = pixel;
}

void setFGObjSetOnTop(Uint32 obj, bool ontop) {
	if(obj > fgobjcnt) {
		DIE(ERROR_INDEX_INVALID, "fgobj");
	}
	fgObjs[obj].isOnTop = ontop;
}

void setFGObjSetElevator(Uint32 obj, bool elevator) {
	if(obj > fgobjcnt) {
		DIE(ERROR_INDEX_INVALID, "fgobj");
	}
	fgObjs[obj].isElevator = elevator;
}

void setFGObjSetPos(Uint32 obj, Uint32 x, Uint32 y) {
	if(obj > fgobjcnt) {
		DIE(ERROR_INDEX_INVALID, "fgobj");
	}
	fgObjs[obj].x = x;
	// Remember last known y position
	fgObjs[obj].last_y = fgObjs[obj].y;
	fgObjs[obj].y = y;
	
	if(abs((int)(fgObjs[obj].last_y - fgObjs[obj].y)) > ELEVATOR_STICKYRANGE) {
		// Object moved to fast to be sticky, so forget last y pos
		fgObjs[obj].last_y = fgObjs[obj].y;
	}
	
}

void setFGObjSetGFX(Uint32 obj, Uint32 gfx) {
	if(obj > fgobjcnt) {
		DIE(ERROR_INDEX_INVALID, "fgobj");
	}
	if(gfx > fgobjgfxcnt) {
		DIE(ERROR_INDEX_INVALID, "gfxobj");
	}	
	fgObjs[obj].gfxobj = gfx;
}


void paintSingleFGObj(SDL_Surface *gfx, Sint32 x, Sint32 y) {
	
	SDL_Rect dest;
	dest.x = x;
	dest.y = y;
	dest.w = gfx->w;
    dest.h = gfx->h;

	// !!! During BLIT, surfaces are MUST NOT BE LOCKED !!!
	// Unlock Surface if needed
	if (SDL_MUSTLOCK(gScreen)) 
		SDL_UnlockSurface(gScreen);

	SDL_BlitSurface(gfx, 0, gScreen, &dest);
	
	// Re-lock screen after BLIT
	if (SDL_MUSTLOCK(gScreen))
		if (SDL_LockSurface(gScreen) < 0) 
			return;	
	
	return;
}

void paintFGObjs(Uint32 xoffs, Uint32 yoffs, bool topObjects) {
	Sint32 xpos;
	Sint32 ypos;
	for(Uint32 i = 0; i < fgobjcnt; i++) {
		xpos =  fgObjs[i].x - xoffs;
		ypos =  fgObjs[i].y - yoffs;
		if(fgObjs[i].isVisible && fgObjs[i].isOnTop == topObjects && xpos + FGOBJ_Surface[fgObjs[i].gfxobj]->w > 0 && ypos + FGOBJ_Surface[fgObjs[i].gfxobj]->h && xpos < SCR_WIDTH && ypos < SCR_HEIGHT) {
			paintSingleFGObj(FGOBJ_Surface[fgObjs[i].gfxobj], xpos, ypos);
		}
	}
}

void doFGPlayerAction(Uint32 playerx, Uint32 playery) {
	Uint32 playerxmax = playerx + TILESIZE;
	Uint32 playerymax = playery + TILESIZE;
	for(Uint32 i = 0; i < fgobjcnt; i++) {
		if(!fgObjs[i].isVisible) {
			continue;
		}
		Uint32 xpos =  fgObjs[i].x;
		Uint32 ypos =  fgObjs[i].y;
		Uint32 xmax = FGOBJ_Surface[fgObjs[i].gfxobj]->w + xpos;
		Uint32 ymax = FGOBJ_Surface[fgObjs[i].gfxobj]->h + ypos;
		
		if(playerxmax > xpos && playerx < xmax && playerymax > ypos && playery < ymax) {
			blLuaCall("handleAction", "i", i);	
		}
	}
}

void handlePlayerFGColission() {
	handlePlayerFGPixelCollission();
	handlePlayerFGTrueCollission();
}

void handlePlayerFGTrueCollission() {
    Sint32 playerx = (Sint32)spritex;
	Sint32 playery = (Sint32)spritey;
	
	// FIXME: Needs special handling for moving elevators!!!!!!!!!!!!!!!!!!!
	
	for(Uint32 i = 0; i < fgobjcnt; i++) {

		if(fgObjs[i].isKilling || !fgObjs[i].isVisible || !fgObjs[i].isBlocking) {
			continue;
		}
		// Check if player CAN touch the object
		if((playerx + TILESIZE) <= fgObjs[i].x || (playery + TILESIZE) <= fgObjs[i].y ||
		      (fgObjs[i].x + FGOBJ_Surface[fgObjs[i].gfxobj]->w) <= playerx ||
		      (fgObjs[i].y + FGOBJ_Surface[fgObjs[i].gfxobj]->h) <= playery) {
            // not overlapping
            continue;
        }
        if(i == 5) {
            printf("TEST\n");
        }
        if(checkPlayerFGTrueCollission(i)) {
            // Check bottom
            if(spritevy > 0.0 && fgOverlap.bottom) {
                spritevy = 0;
                spritey -= fgOverlap.bottom;
                isJumping = false;
                checkPlayerFGTrueCollission(i);
            }
            // Check left
            if(spritevx < 0.0 && fgOverlap.left) {
                spritevx = 0;
                spritex += fgOverlap.left;
                checkPlayerFGTrueCollission(i);
            }
            // Check right
            if(spritevx > 0.0 && fgOverlap.right > 0) {
                spritevx = 0;
                spritex -= fgOverlap.right;
                checkPlayerFGTrueCollission(i);
            }
            // Check top
            if(spritevy < 0.0 && fgOverlap.top) {
                spritevy = 0;
                spritey += fgOverlap.top;
                //checkPlayerFGTrueCollission(i);
            }

        }
    }
    
}


bool getPlayerFGKillColission() {
	Sint32 playerx = (Sint32)spritex;
	Sint32 playery = (Sint32)spritey;
	
	for(Uint32 i = 0; i < fgobjcnt; i++) {
		if(!fgObjs[i].isKilling || !fgObjs[i].isVisible) {
			continue;
		}
		
		if((playerx + TILESIZE) > fgObjs[i].x && playerx < (fgObjs[i].x + FGOBJ_Surface[fgObjs[i].gfxobj]->w) &&
		   (playery + TILESIZE) > fgObjs[i].y && playery < (fgObjs[i].y + FGOBJ_Surface[fgObjs[i].gfxobj]->h) &&
           checkPlayerFGTrueCollission(i)) {
			return true;
		}
	}
	return false;
}

void handlePlayerFGPixelCollission() {
	Sint32 playerx = (Sint32)spritex;
	Sint32 playery = (Sint32)spritey;
	
	for(Uint32 i = 0; i < fgobjcnt; i++) {
		if(!fgObjs[i].isPixel || !fgObjs[i].isVisible) {
			continue;
		}
		
		if((playerx + TILESIZE) > fgObjs[i].x && playerx < (fgObjs[i].x + FGOBJ_Surface[fgObjs[i].gfxobj]->w) &&
		   (playery + TILESIZE) > fgObjs[i].y && playery < (fgObjs[i].y + FGOBJ_Surface[fgObjs[i].gfxobj]->h) &&
           checkPlayerFGTrueCollission(i)) {
			foundPixels++;
			gamedata.player->score += PIXEL_SCORE;
			soundPlayFX(FX_COLLECT_PIXEL);
			fgObjs[i].isVisible = false;
		}
	}
}

bool checkPlayerFGTrueCollission(Uint32 objid) {

	bool overlap = false;
	fgOverlap.top = 0;
	fgOverlap.bottom = 0;
	fgOverlap.left = 0;
	fgOverlap.right = 0;



	SDL_Surface* obj = FGOBJ_Surface[fgObjs[objid].gfxobj];
	
	if (SDL_MUSTLOCK(sprite))
		if (SDL_LockSurface(sprite) < 0) 
			return false;
	if (SDL_MUSTLOCK(obj))
		if (SDL_LockSurface(obj) < 0) 
			return false;
	
	Uint32 pitch_sprite = sprite->pitch / 4;
	Uint32 pitch_obj = obj->pitch / 4;
	Sint32 i, j, k, l;
	Uint32 pixel;
	
	Sint32 playerscreenx, playerscreeny;
	Sint32 objx, objy;
	Sint32 minx = -1;
	Sint32 miny = -1;
	Sint32 maxx = TILESIZE;
	Sint32 maxy = TILESIZE;
	
	// Copy the sprite into the screen
	for(i=0; i < TILESIZE; i++) {
		for(j=0; j < TILESIZE; j++) {
			pixel = ((Uint32 *)sprite->pixels)[(j + playersprite_srcoffset) * pitch_sprite + i];
			
			if(pixel != 0x0000ff00) {
				playerscreenx = (Uint32)spritex + i;
				playerscreeny = (Uint32)spritey + j;
				// Check, if this point CAN touch the sprite...
				if(playerscreenx >= fgObjs[objid].x && playerscreeny >= fgObjs[objid].y &&
				   playerscreenx < (fgObjs[objid].x + obj->w) &&
				   playerscreeny < (fgObjs[objid].y + obj->h)) {
					
					// OK, pixel overlaps, calculate in-object position and check for transparency
					objx = playerscreenx - fgObjs[objid].x;
					objy = playerscreeny - fgObjs[objid].y;
					
					pixel = ((Uint32 *)obj->pixels)[objy * pitch_obj + objx];
					if((pixel & 0xff000000) > 0x80000000) {
						overlap = true;
						// Search to left
						if(i < TILESIZE/2) {
						     minx = TILESIZE;
						     for(k = i; k >= 0; k--) {
                                 if(((Uint32 *)sprite->pixels)[(j + playersprite_srcoffset) * pitch_sprite + k] != 0x0000ff00) {
                                    minx = k;
                                }
                            }
                            minx = i - minx + 1;
                            if(minx > fgOverlap.left) {
                                fgOverlap.left = minx;
                            }
                        }

                        // Search to right
						if(i >= TILESIZE/2) {
						     maxx = 0;
						     for(k = i; k < TILESIZE; k++) {
                                 if(((Uint32 *)sprite->pixels)[(j + playersprite_srcoffset) * pitch_sprite + k] != 0x0000ff00) {
                                    maxx = k;
                                }
                            }
                            maxx = maxx - i + 1;
                            if(maxx > fgOverlap.right) {
                                fgOverlap.right = maxx;
                            }
                        }

                        // Search to top
						if(j < TILESIZE/2) {
						     miny = TILESIZE;
						     for(l = j; l >= 0; l--) {
                                 if(((Uint32 *)sprite->pixels)[(l + playersprite_srcoffset) * pitch_sprite + i] != 0x0000ff00) {
                                    miny = l;
                                }
                            }
                            miny = j - miny + 1;
                            if(miny > fgOverlap.top) {
                                fgOverlap.top = miny;
                            }
                        }

                        // Search to bottom
						if(j >= TILESIZE/2) {
						     maxy = 0;
						     for(l = j; l < TILESIZE; l++) {
                                 if(((Uint32 *)sprite->pixels)[(l + playersprite_srcoffset) * pitch_sprite + i] != 0x0000ff00) {
                                    maxy = l;
                                }
                            }
                            maxy = maxy - j + 1;
                            if(maxy > fgOverlap.bottom) {
                                fgOverlap.bottom = maxy;
                            }
                        }
					}
				}
				//((Uint32 *)gScreen->pixels)[(j + y) * pitch_bg + i + x] = pixel;
			}
			
		}
	}
	
	// More than a half-overlap shouldn't be possible; but check anyway...
	if(fgOverlap.left > TILESIZE/2) {
		fgOverlap.left = 0;
	}
	if(fgOverlap.top > TILESIZE/2) {
		fgOverlap.top = 0;
	}
	if(fgOverlap.right > TILESIZE/2) {
		fgOverlap.right = 0;
	}
	if(fgOverlap.bottom > TILESIZE/2) {
		fgOverlap.bottom = 0;
	}
	if(!fgOverlap.left && !fgOverlap.right && !fgOverlap.top && !fgOverlap.bottom) {
        // Someone set up us the bomb????
        overlap = false;
    }
	
	if(overlap) {
		//printf("L: %d R: %d T: %d B: %d\n", fgOverlap.left, fgOverlap.right, fgOverlap.top, fgOverlap.bottom);
		
	}
		// Unlock Surfaces if needed
		if (SDL_MUSTLOCK(sprite)) 
			SDL_UnlockSurface(sprite);
		if (SDL_MUSTLOCK(obj)) 
			SDL_UnlockSurface(obj);
		
	return overlap;
}
