// BlinkGame - Hunt for the Lost Pixels
// an SDL game in honor of the BlinkenArea crew
//
// (C) 2005-07 Rene Schickbauer, Wolfgang Dautermann
//
// See LICENSE for licensing information
//


#include <stdlib.h>
#include <string.h>
#include "globals.h"
#include "sound.h"
#include "engine.h"
#include "showtext.h"
#include "debug.h"
#include "levelhandler.h"
#include "showloading.h"
#include "gameengine.h"
#include "joystick.h"
#include "cmdopts.h"
#include "errorhandler.h"
#include "fonthandler.h"
#include "menu.h"
#include "showvideo.h"
#include "minilzo.h"
#include "extractmetabmf.h"
#include "fginlay.h"
#include "intro.h"
#ifdef HASOGGSUPPORT
#include "oggplayer.h"
#endif // HASOGGSUPPORT
#include "splash.h"

#ifndef DISABLE_NETWORK
#include <SDL_net.h>
#endif // DISABLE_NETWORK



// Screen surface
SDL_Surface *gScreen = 0;

// Last iteration's tick value
#include <SDL_types.h>
Uint32 gLastTick;

#ifdef ENABLE_FULLSCREEN
#ifdef START_WITH_FULLSCREEN
bool isFullscreen = true;
#else // NO START_WITH_FULLSCREEN
bool isFullscreen = false;
#endif // START_WITH_FULLSCREEN
#endif // ENABLE_FULLSCREEN
#ifdef DISABLE_SOUND
bool soundOK = false;
#else // DISABLE_SOUND
bool soundOK = true;
#endif // DISABLE_SOUND


/* global variable. declared as extern in cmdopts.cpp */
CMD_OPTS cmd_opt_data[] = {
/* {short option name, long option name, description, is_boolean, defaultvalue_boolean, defaultvalue_string} */
#ifdef ENABLE_FULLSCREEN
#ifdef START_WITH_FULLSCREEN
{"-f", "--fullscreen", "Use Fullscreen (default)", true, 0, NULL},
{"-w", "--windowed", "Use windowed mode", true, 0, NULL},
#else // START_WITH_FULLSCREEN not defined
{"-f", "--fullscreen", "Use Fullscreen", true, 0, NULL},
{"-w", "--windowed", "Use windowed mode (default)", true, 0, NULL},
#endif // START_WITH_FULLSCREEN
#endif // ENABLE_FULLSCREEN
#ifndef DISABLE_SOUND
{"-s", "--disable-sound", "Disable sound", true, 0, NULL},
#endif // DISABLE_SOUND
{"-v", "--version", "Print version information and exit", true, 0, NULL},
{"-h", "--help", "Command-line help", true, 0, NULL},
#ifndef DISABLE_NETWORK
{"-p", "--proxyurl", "URL of the HTTP Proxy (--proxyurl=http://proxy.example.com:port or -p=http://proxy.example.com:port)", false, 0, NULL},
#endif // DISABLE_NETWORK

/* currently not supported, therefore commented
{"-r", "--respath", "Set path to game data", false, 0, NULL}, 
*/
CMD_OPTS_END /* {"END", "END", "END", false, 0, "END"} */
};

#ifndef DISABLE_NETWORK
char * proxyurl ; 
#endif // DISABLE_NETWORK

// Entry point
int main(int ARGC, char **ARGV)
{
	int numoptions ;
	
	// Basic configuration init
	configInit();

	// Initialize the mini-lzo we use for BMF handling
    printf("\nProgram is using LZO real-time data compression library (v%s, %s).\n",
            lzo_version_string(), lzo_version_date());
    printf("Copyright (C) 1996-2005 Markus Franz Xaver Johannes Oberhumer\nAll Rights Reserved.\n\n");
    // We use example code from libtheora (BSD License) */
    printf("\nIncludes code from libtheora Copyright (C) Xiph.org Foundation.\n\n");
    if (lzo_init() != LZO_E_OK)
    {
		DIE(ERROR_LZO, "lzo_init()");
    }
	
	// Parse command-line options
	if(!initCmdOpt(ARGC, ARGV)) {
		helpCmdOpt(ARGV[0]); 
		exit(1);
	}
	numoptions = getCmdOptBool("-h");
	if (numoptions < 0) { helpCmdOpt(ARGV[0]) ; exit(1); } /* Error */
	if (numoptions > 0) { helpCmdOpt(ARGV[0]) ; exit(0); }
	
#ifdef ENABLE_FULLSCREEN
	numoptions = getCmdOptBool("-f");
	if (numoptions < 0) { helpCmdOpt(ARGV[0]) ; exit(1); } /* Error */
	if (numoptions > 0) { isFullscreen = true; }

	numoptions = getCmdOptBool("-w");
	if (numoptions < 0) { helpCmdOpt(ARGV[0]) ; exit(1); } /* Error */
	if (numoptions > 0) { isFullscreen = false; }
#endif // ENABLE_FULLSCREEN

#ifndef DISABLE_SOUND
	numoptions = getCmdOptBool("-s");
	if (numoptions < 0) { helpCmdOpt(ARGV[0]) ; exit(1); } /* Error */
	if (numoptions > 0) { soundOK = false; }
#endif // DISABLE_SOUND

#ifndef DISABLE_NETWORK
	proxyurl = getCmdOptString("-p");
	if (proxyurl != NULL ) {
		if (strlen(proxyurl) >=255) {  /* NOTE: Probably use a errormacro or something 	similiar? */
			fprintf(stderr,"Error. The Proxyurl must be <= 255 Characters. \n"); /* Defined in RFC 2181: A FQDN is limited to 255 chars */
			exit(1);
		} 
	} ; 

#endif // DISABLE_NETWORK


	numoptions = getCmdOptBool("-v");
	if (numoptions < 0) { exit(1); } /* Error */
	if (numoptions > 0) {
		printf("BlinkenSisters - Hunt for the Lost Pixels\nLostPixels %s \n",LONGVERSION);
		printf("\nCompile-Flags:\n");
#ifdef ENABLE_FULLSCREEN
		printf("  Fullscreen: ENABLED\n");
#else // NO ENABLE_FULLSCREEN
		printf("  Fullscreen: DISABLED\n");
#endif // ENABLE_FULLSCREEN
#ifndef DISABLE_SOUND
		printf("  Sound-Support: ENABLED\n");
#else // NO DISABLE_SOUND
		printf("  Sound-Support: DISABLED\n");
#endif // DISABLE_SOUND
#ifndef DISABLE_NETWORK
		printf("  Networking-Support: ENABLED\n");
#else // NO DISABLE_NETWORK
		printf("  Networking-Support: DISABLED\n");
#endif // DISABLE_NETWORK
		printf("  Resource-Path: %s\n", configGetPath(""));
		exit(0);
	}	
	// END processing command-line arguments
	
	// Initialize SDL's subsystems
	Uint32 SDLinitflags = SDL_INIT_VIDEO | SDL_INIT_JOYSTICK ; 
	if (soundOK!=false) { /* sound is compiled and not disabled by command line switch */
		SDLinitflags = SDLinitflags | SDL_INIT_AUDIO ;
	}
#ifndef CAVAC_RELEASEMODE
		SDLinitflags = SDLinitflags |SDL_INIT_NOPARACHUTE;
#endif // CAVAC_RELEASEMODE

	if ( SDL_Init(SDLinitflags) < 0 ) {
		fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
		exit(1);
	}

#ifndef DISABLE_NETWORK
	if ( SDLNet_Init() < 0 ) {
		fprintf(stderr, "Unable to init SDLNet: %s\n", SDLNet_GetError());
		exit(1);
	}
#endif // DISABLE_NETWORK


	// Initialize sound subsystem
	initSound();
	initJoystick();
	initFontHandler();
	initShowVideo();
	
	// Register SDL_Quit to be called at exit; makes sure things are
	// cleaned up when we quit.
	//atexit(deInitJoystick);
	atexit(deInitExtractMetaBMF);
	atexit(deInitShowVideo);
	atexit(deInitFontHandler);
	atexit(deInitSound);
	atexit(SDL_Quit);


	// Show splash screen
	showSplash();
	
	// Attempt to create a SCR_WIDTHxSCR_HEIGHT window with 32bit pixels.
	
#ifdef ENABLE_FULLSCREEN
	if(isFullscreen) {
		gScreen = SDL_SetVideoMode(SCR_WIDTH, SCR_HEIGHT, 32, SDL_SWSURFACE | SDL_FULLSCREEN);
	} else {
		gScreen = SDL_SetVideoMode(SCR_WIDTH, SCR_HEIGHT, 32, SDL_SWSURFACE);
	}
#else // NO ENABLE_FULLSCREEN
	gScreen = SDL_SetVideoMode(SCR_WIDTH, SCR_HEIGHT, 32, SDL_SWSURFACE);
#endif // ENABLE_FULLSCREEN

	// If we fail, return error.
	if ( gScreen == NULL ) 
	{
		fprintf(stderr, "Unable to set %d x %d videomode: %s\n", SCR_WIDTH, SCR_HEIGHT, SDL_GetError());
		exit(1);
	}


	// Hide the mousepointer
	SDL_ShowCursor(SDL_DISABLE);
	/* Set a window title. */
	SDL_WM_SetCaption("Blinkensisters - Hunt for the lost pixels", NULL); // Window title: blinkensisters, Icon title: none
	SDL_WM_SetIcon(IMG_Load(configGetPath("lostpixels.png")), NULL);
	
	
	// Initialize loader screen and show it
	initShowLoading("loading.jpg");
	showLoading();
	initExtractMetaBMF();
	initFGInlay();

	// Inits that need an initialized screen
	initShowText();
	initFPSCounter();
	initMenu();

	// Display the Intro sequence
#ifdef CAVAC_RELEASEMODE
#ifdef HASOGGSUPPORT
	playogg(configGetPath("startanim_video.ogg"),gScreen); /* play the standard ogg video instead of the BMF showVideo("startanim.bmf", true); */
#else // HASOGGSUPPORT	
	showVideo("startanim.bmf", true, true);
#endif // NO HASOGGSUPPORT
#endif // CAVAC_RELEASEMODE
	configStartupComplete();	

#ifndef BLINKOMATEDITION	
	while(menuDisplay()) {
		//
	}
#else // Running on Blinkomat
	while(true) {
		menuAttrackMode();
		menuAddonDisplay();
	}
#endif //BLINKOMATEDITION

						  
						  
// Clean up
	deInitMenu();
	deInitShowLoading();
	deInitFGInlay();
	SDL_ShowCursor(SDL_ENABLE);
	return 0;
}


#if WIN32

#include <windows.h>
#include <stdio.h>
#include <fcntl.h>
#include <io.h>
#include <fstream>

void RedirectIOToConsole()
{
	int hConHandle;
	long lStdHandle;
	CONSOLE_SCREEN_BUFFER_INFO coninfo;
	FILE *fp;

	// allocate a console for this app

	AllocConsole();
	// set the screen buffer to be big enough to let us scroll text

	GetConsoleScreenBufferInfo(GetStdHandle(STD_OUTPUT_HANDLE),&coninfo);
	coninfo.dwSize.Y = 40;
	SetConsoleScreenBufferSize(GetStdHandle(STD_OUTPUT_HANDLE),	coninfo.dwSize);

	// redirect unbuffered STDOUT to the console

	lStdHandle = (long)GetStdHandle(STD_OUTPUT_HANDLE);
	hConHandle = _open_osfhandle(lStdHandle, _O_TEXT);

	fp = _fdopen( hConHandle, "w" );

	*stdout = *fp;
	setvbuf( stdout, NULL, _IONBF, 0 );

	// redirect unbuffered STDIN to the console
	lStdHandle = (long)GetStdHandle(STD_INPUT_HANDLE);
	hConHandle = _open_osfhandle(lStdHandle, _O_TEXT);

	fp = _fdopen( hConHandle, "r" );
	*stdin = *fp;
	setvbuf( stdin, NULL, _IONBF, 0 );

	// redirect unbuffered STDERR to the console

	lStdHandle = (long)GetStdHandle(STD_ERROR_HANDLE);
	hConHandle = _open_osfhandle(lStdHandle, _O_TEXT);
	fp = _fdopen( hConHandle, "w" );
	*stderr = *fp;
	setvbuf( stderr, NULL, _IONBF, 0 );
}

void RedirectIOToFiles()
{
	int hConHandle;
	long lStdHandle;
	FILE *fp;

	// allocate a console for this app


	// redirect unbuffered STDOUT to the console

	fp = fopen( "stdoutstderr_lp.txt", "w" );

	*stdout = *fp;
	setvbuf( stdout, NULL, _IONBF, 0 );

	// redirect unbuffered STDERR to the console
	*stderr = *fp;
	setvbuf( stderr, NULL, _IONBF, 0 );
}


int WINAPI WinMain(  HINSTANCE hInstance,      // handle to current instance
  HINSTANCE hPrevInstance,						// handle to previous instance
  LPSTR lpCmdLine,								// command line
  int nCmdShow									// show state
  )
{
	int argc = 0;
	int i=0;
	if (lpCmdLine[0]) // at least one paramteter
		argc = 1;
	while (lpCmdLine[i])
	{
		if (lpCmdLine[i] == ' ')
			argc++;
		i++;
	}
	argc++;  // add app name

	char** argv = (char**)malloc(argc*sizeof(char*));
	char appName[500];
	GetModuleFileNameA(0, appName, sizeof(appName));
	char* p;
	
	while(p= strrchr(appName,'\\'))
		*p= '/';
	p= argv[0] = (char*)malloc(strlen(appName)+10);
	strcpy(p,appName);

	i = argc-1;
	char* strptr = strrchr(lpCmdLine, ' ');
	while (strptr)
	{
		p=argv[i] = (char*)malloc(strlen(strptr+1)+10);
		strcpy( p,strptr+1);
		*strptr = '\0';
		i--;
		strptr = strrchr(lpCmdLine, ' ');
	}
	if (lpCmdLine[0])
	{
		p=argv[1] = (char*)malloc(strlen(lpCmdLine)+1);
		strcpy(p,lpCmdLine);
	}
	//RedirectIOToConsole();
	RedirectIOToFiles();

	return main(argc, argv);
}
#endif // WIN32


